
/* S2 cpu simulator

   Computer Architecture 1997, 1998

   simplify (direct) for calculator project   29 Jan 2016
   no Z, S
   add address trace (Chinese new year)		  10 Feb 2016
   add interrupt 							  27 Mar 2016
   clean up for "grandfather s21" source      14 Jan 2017

   P. Chongstitvatana
   Department of Computer Engineering
   Chulalongkorn University

*/

#include "s21.h"

int R[32], Ir, Pc, M[MAXMEM], RetAds, intflag;
int runflag, savePc, lastInt;
extern char *cp2;		// output buffer

int signx2( int d ){	// sign bit 21 extended
	if( d & 0x0200000 ) return d | 0xFFC00000;
	return d;
}

void trap( int reg, int num ){	// special function

	switch( num ) {
	case 0:  // stop
		printf("\nstop, execute %d instructions\n",ninst);
		runflag = 0;
		break;
	case 1: // print integer
		printf("%d",R[reg]);
		break;
	case 2: // printc
		printf("%c",R[reg]);
		break;
	}
}

/*
void interrupt(void){
	if( intflag ){
		printf("interrupt\n");
		lastInt = ninst;	// reset int interval
		RetAds = Pc;		// save pc
		Pc = M[INTVEC];
		if( Pc < 0 || Pc > 10000)
			error("Pc out of range");
	}
}
*/
void run(void){		// execute one instruction
	int ads, d, r1, r2, r3, i;

	savePc = Pc;
	Ir = M[Pc];		// instruction fetch
	Pc++;
	ads = IRads();	// decode
	d = IRdisp();	// signx  17-bit
	r1 = IRr1();
	r2 = IRr2();
	r3 = IRr3();
	ninst++;
	switch( IRop() ) {
	case NOP:  break;
	case LDA:  R[r1] = M[ads];  break;
	case LDD:  R[r1] = M[R[r2]+ d];  break;
	case STA:  M[ads] = R[r1];   break;
	case STD:  M[R[r2] + d] = R[r1];  break;
	case MVI:  R[r1] = signx2(ads); break;
	case JMP:  Pc = ads; break;
	case JAL:  R[r1] = Pc; Pc = ads; break;
	case JT: if( R[r1] != 0 ) Pc = ads; break;
	case JF: if( R[r1] == 0 ) Pc = ads; break;
	case ADDI: R[r1] = R[r2] + d; break;
	case SUBI: R[r1] = R[r2] - d; break;
	case MULI: R[r1] = R[r2] * d; break;
	case DIVI: R[r1] = R[r2] / d; break;
	case ANDI: R[r1] = R[r2] & d; break;
	case ORI:  R[r1] = R[r2] | d; break;
	case XORI: R[r1] = R[r2] ^ d; break;
	case EQI:  R[r1] = R[r2] == d; break;
	case NEI:  R[r1] = R[r2] != d; break;
	case LTI:  R[r1] = R[r2] <  d; break;
	case LEI:  R[r1] = R[r2] <= d; break;
	case GTI:  R[r1] = R[r2] >  d; break;
	case GEI:  R[r1] = R[r2] >= d; break;
	case SHLI: R[r1] = R[r2] << d; break;
	case SHRI: R[r1] = R[r2] >> d; break;
	case MODI: R[r1] = R[r2] %  d; break;
	case XOP:
		switch( IRxop() ){
		case ADD:  R[r1] = R[r2] + R[r3]; break;
		case SUB:  R[r1] = R[r2] - R[r3]; break;
		case MUL:  R[r1] = R[r2] * R[r3]; break;
		case DIV:  R[r1] = R[r2] / R[r3]; break;
		case AND:  R[r1] = R[r2] & R[r3]; break;
		case OR:   R[r1] = R[r2] | R[r3]; break;
		case XOR:  R[r1] = R[r2] ^ R[r3]; break;
		case EQ:   R[r1] = R[r2] == R[r3]; break;
		case NE:   R[r1] = R[r2] != R[r3]; break;
		case LT:   R[r1] = R[r2] <  R[r3]; break;
		case LE:   R[r1] = R[r2] <= R[r3]; break;
		case GT:   R[r1] = R[r2] >  R[r3]; break;
		case GE:   R[r1] = R[r2] >= R[r3]; break;
		case SHL:  R[r1] = R[r2] << R[r3]; break;
		case SHR:  R[r1] = R[r2] >> R[r3]; break;
		case MOD:  R[r1] = R[r2] %  R[r3]; break;
		case MOV:  R[r1] = R[r2]; break;
		case LDX:  R[r1] = M[R[r2] + R[r3]]; break;
		case STX:  M[R[r2] + R[r3]] = R[r1]; break;
		case RET:  Pc = R[r1]; break;
		case TRAP: trap(r1,r2); break;
		case PUSH:
			R[r1]++;			// increment stack pointer
			M[R[r1]] = R[r2];
			break;
		case POP:
			R[r2] = M[R[r1]];
			R[r1]--;			// decrement stack pointer
			break;
		case NOT:  R[r1] = (R[r2] == 0) ? ~0 : 0;  break;

		default:
			error("undefine xop");
		}
		break;
	default:
		error("undefine op");
	}
}

int main(int argc, char *argv[]){
	if( argc < 2 ) {
		printf("usage : s21 objfile\n");
		exit(0);
	}
	loadprogram(argv[1]);

	clock = 0;
	ninst = 0;
	Pc = 0;
	runflag = 1;
	intflag = 1;
	lastInt = 0;
	interp();

	return 0;
}

